from ednaml.trainer.BaseTrainer import BaseTrainer

from ednaml.trainer.ClassificationTrainer import ClassificationTrainer
from ednaml.trainer.MultiClassificationTrainer import MultiClassificationTrainer
from ednaml.trainer.MultiBranchTrainer import MultiBranchTrainer
