# Proposed Splits for SUN from
# Zero-shot Learning - The Good, the Bad and the Ugly. Y. Xian, B. Schiele, Z. Akata. IEEE CVPR 2017.

trainval = r"""bakery_shop
bar
barn
boathouse
bookstore
bridge
butchers_shop
candy_store
carrousel
castle
cliff
closet
coast
dam
dock
fastfood_restaurant
fly_bridge
fountain
home_theater
house
lift_bridge
lighthouse
manufactured_home
mobile_home
nursery
palace
patio
pier
restaurant
restaurant_patio
roof
rope_bridge
sandbar
sawmill
sea_cliff
seawall
roundabout
toyshop
trestle_bridge
underwater_coral_reef
underwater_wreck
valley
viaduct
volcano
water_tower
vegetable_garden
chemical_plant
residential_neighborhood
tent_indoor
utility_room
swimming_hole
darkroom
shipping_room
shed
gangplank
forest_broadleaf
ferryboat_outdoor
campus
railroad_track
auto_racing_paddock
zoo
garage_outdoor
strip_mine
fish_farm
airfield
glacier
dolmen
bistro_outdoor
music_store
diner_outdoor
general_store_outdoor
bullpen
barndoor
discotheque
poolroom_home
warehouse_indoor
art_studio
thriftshop
campsite
golf_course
jail_indoor
train_railway
carport_outdoor
atrium_public
museum_outdoor
woodland
fire_escape
fjord
brewery_outdoor
sun_deck
balcony_interior
videostore
underwater_kelp_forest
ski_lodge
cybercafe
tent_outdoor
flea_market_indoor
day_care_center
village
mesa
mastaba
subway_interior
jury_box
earth_fissure
herb_garden
chalet
beer_garden
conference_center
wharf
food_court
drill_rig
throne_room
hunting_lodge_outdoor
control_room
dinette_home
military_hut
berth
kitchen
water_treatment_plant_outdoor
baggage_claim
fence
clothing_store
library_outdoor
mission
revolving_door
sand_trap
amphitheater
hill
aquatic_theater
cathedral_indoor
nursing_home
tea_garden
editing_room
dinette_vehicle
gazebo_exterior
limousine_interior
forest_needleleaf
basketball_court_outdoor
martial_arts_gym
brickyard_outdoor
donjon
trench
particle_accelerator
bus_shelter
clean_room
banquet_hall
weighbridge
caravansary
rainforest
recycling_plant_outdoor
newsroom
assembly_line
flood
elevator_door
atrium_home
oilrig
call_center
estuary
pulpit
office_cubicles
checkout_counter
van_interior
ski_slope
apse_outdoor
morgue
ice_skating_rink_indoor
chicken_farm_indoor
raceway
mini_golf_course_outdoor
shower
stone_circle
subway_station_corridor
market_outdoor
velodrome_outdoor
pharmacy
power_plant_indoor
canteen
dining_hall
desert_vegetation
overpass
drainage_ditch
print_shop
circus_tent_outdoor
steel_mill_indoor
bank_indoor
butte
kindergarden_classroom
ramp
waterfall_cascade
military_hospital
rolling_mill
conference_hall
baseball_field
boat_deck
subway_station_platform
control_tower_indoor
schoolyard
bus_station_outdoor
snowfield
anechoic_chamber
watchtower
formal_garden
sewing_room
catwalk
burial_chamber
alcove
carport_freestanding
airport_ticket_counter
embassy
office
stable
liquor_store_indoor
arena_performance
raft
kiosk_outdoor
resort
cabin_outdoor
pizzeria
dugout
courthouse
movie_theater_indoor
physics_laboratory
cardroom
tennis_court_outdoor
stage_indoor
badlands
lake_natural
waiting_room
pitchers_mound
pantry
lido_deck_outdoor
shipyard_outdoor
escalator_outdoor
crawl_space
wine_cellar_barrel_storage
cathedral_outdoor
temple_east_asia
theater_indoor_procenium
cottage
dry_dock
hallway
outcropping
parking_garage_indoor
ice_skating_rink_outdoor
phone_booth
ghost_town
factory_outdoor
bedchamber
botanical_garden
chicken_coop_outdoor
corn_field
pavilion
firing_range_outdoor
bamboo_forest
industrial_area
pig_farm
gatehouse
restroom_outdoor
construction_site
gasworks
dacha
nuclear_power_plant_indoor
field_road
lab_classroom
attic
landfill
driving_range_outdoor
fire_station
corridor
restaurant_kitchen
underwater_ocean_deep
witness_stand
aquarium
manhole
courtroom
poolroom_establishment
hotel_outdoor
bedroom
kasbah
basilica
skatepark
joss_house
nuclear_power_plant_outdoor
jewelry_shop
tunnel_rail_outdoor
mansion
bakery_kitchen
plantation_house
library_indoor
basketball_court_indoor
oast_house
cloister_outdoor
underwater_pool
tower
courtyard
portico
freeway
gift_shop
childs_room
cloister_indoor
quay
gulch
bell_foundry
moat_water
conference_room
driveway
grotto
pagoda
television_studio
escalator_indoor
pilothouse_outdoor
theater_indoor_round
ice_floe
levee
art_school
booth_indoor
control_tower_outdoor
office_building
hut
swamp
liquor_store_outdoor
badminton_court_outdoor
auto_showroom
oasis
plaza
inn_outdoor
youth_hostel
canal_urban
veranda
synagogue_indoor
ski_jump
desert_sand
park
waterfall_cataract
guardhouse
repair_shop
pond
factory_indoor
pump_room
hedge_maze
restroom_indoor
beer_hall
beach_house
riding_arena
podium_indoor
balcony_exterior
field_wild
soccer_field
organ_loft_exterior
desert_road
hot_spring
airport_airport
hospital_room
topiary_garden
hot_tub_indoor
podium_outdoor
badminton_court_indoor
science_museum
farm
amusement_arcade
cafeteria
downtown
baptistry_indoor
parlor
heliport
garbage_dump
theater_outdoor
train_station_outdoor
gymnasium_indoor
packaging_plant
ball_pit
fortress
bowling_alley
sky
heath
airport_terminal
lagoon
cargo_deck_airplane
mine
power_plant_outdoor
mausoleum
aqueduct
reading_room
quonset_hut_outdoor
computer_room
kennel_outdoor
catacomb
waterfall_plunge
junkyard
tree_house
labyrinth_indoor
highway
bleachers_outdoor
game_room
quadrangle
water_mill
lobby
confessional
salt_plain
shoe_shop
mountain_path
airport_entrance
cabana
apartment_building_outdoor
access_road
batting_cage_indoor
mosque_outdoor
airplane_cabin
putting_green
kennel_indoor
archaelogical_excavation
natural_history_museum
recreation_room
submarine_interior
window_seat
teashop
shopping_mall_indoor
railway_yard
wine_cellar_bottle_storage
hacienda
living_room
chaparral
movie_theater_outdoor
pasture
hospital
casino_indoor
elevator_shaft
water_treatment_plant_indoor
shopfront
cavern_indoor
root_cellar
priory
armory
stadium_outdoor
stadium_football
greenhouse_outdoor
bicycle_racks
planetarium_outdoor
gun_deck_indoor
wave
engine_room
boardwalk
bullring
dentists_office
classroom
terrace_farm
drugstore
athletic_field_outdoor
orchard
beach
junk_pile
auto_factory
pub_outdoor
toll_plaza
observatory_indoor
building_facade
swimming_pool_outdoor
locker_room
lecture_room
inn_indoor
car_interior_backseat
stadium_baseball
jail_outdoor
doorway_outdoor
lean-to
veterinarians_office
flea_market_outdoor
baptistry_outdoor
strip_mall
boxing_ring
supermarket
biology_laboratory
ruin
sacristy
beauty_salon
mountain_road
ocean
chicken_coop_indoor
creek
cottage_garden
harbor
hayfield
rice_paddy
pilothouse_indoor
jacuzzi_outdoor
marsh
dining_car
velodrome_indoor
ticket_window_outdoor
home_office
parade_ground
mineshaft
gas_station
sauna
greenhouse_indoor
departure_lounge
pedestrian_overpass_outdoor
winery
pet_shop
zen_garden
arch
forest_road
football_field
batters_box
chicken_farm_outdoor
runway
elevator_lobby
track_indoor
windmill
lake_artificial
apse_indoor
lock_chamber
ranch_house
medina
underwater_ocean_shallow
crevasse
outhouse_outdoor
garage_indoor
skyscraper
amusement_park
wind_farm
recycling_plant_indoor
promenade
moor
chapel
street
kiosk_indoor
slum
bayou
temple_western
porch
river
loading_dock
general_store_indoor
tollbooth
museum_indoor
hedgerow
voting_booth
islet
waterfall_block
machine_shop
foundry_indoor
igloo
irrigation_ditch
staircase
hunting_lodge_indoor
coffee_shop
dairy_indoor
auto_mechanics_indoor
stage_outdoor
dorm_room
circus_tent_indoor
tree_farm
market_indoor
fairway
moat_dry
bus_interior
newsstand_outdoor
optician
tearoom
convenience_store_indoor
road_cut
underwater_ice
jail_cell
crosswalk
iceberg
naval_base
bow_window_outdoor
reception
tennis_court_indoor
track_outdoor
lookout_station_outdoor
hot_tub_outdoor
town_house
picnic_area
arrival_gate_outdoor
arena_hockey
train_depot
covered_bridge_exterior
mountain_snowy
backstage
signal_box
flight_of_stairs_natural
art_gallery
cockpit
great_hall
stilt_house_water
operating_room
diner_indoor
rock_arch
ventilation_shaft
kitchenette
furnace_room
church_outdoor
bistro_indoor
trailer_park
spillway
oil_refinery_outdoor
waterfall_fan
labyrinth_outdoor
playroom
bindery
building_complex
flight_of_stairs_urban
ranch
laundromat
wheat_field
workroom
rubble
mountain
hangar_outdoor
delicatessen
elevator_freight_elevator
basement
schoolhouse
city
fort
abbey
steel_mill_outdoor
wet_bar
squash_court
geodesic_dome_outdoor
parking_garage_outdoor
electrical_substation
sushi_bar
bank_outdoor
planetarium_indoor
florist_shop_indoor
foundry_outdoor
stadium_soccer
funeral_chapel
canyon
music_studio
gorge
hotel_breakfast_area
train_station_station
industrial_park
cheese_factory
dining_room
swimming_pool_indoor
server_room
ossuary
synagogue_outdoor
forest_path
bathroom"""

query = r"""airlock
temple_south_asia
monastery_outdoor
piano_store
watering_hole
geodesic_dome_indoor
chemistry_lab
vineyard
brewery_indoor
pub_indoor
vestry
bog
wrestling_ring_indoor
trading_floor
volleyball_court_outdoor
racecourse
cemetery
betting_shop
casino_outdoor
theater_indoor_seats
dirt_track
motel
ballroom
sandbox
mosque_indoor
batting_cage_outdoor
bow_window_indoor
car_interior_frontseat
alley
nightclub
arena_basketball
hotel_room
jacuzzi_indoor
tunnel_road_outdoor
bank_vault
excavation
lawn
galley
bazaar_outdoor
canal_natural
train_station_platform
cubicle_office
fishpond
ice_shelf
bazaar_indoor
field_cultivated
exhibition_hall
observatory_outdoor
rectory
tundra
firing_range_indoor
doorway_indoor
ticket_booth
bus_depot_outdoor
auditorium
ski_resort
ziggurat
elevator_interior
artists_loft
archive
playground
hoodoo
parking_lot
yard
corral
japanese_garden
promenade_deck
landing_deck
workshop
hangar_indoor
church_indoor
savanna"""

trainval = trainval.split("\n")
query = query.split("\n")
