# Proposed splits for CUB200 from
# Zero-shot Learning - The Good, the Bad and the Ugly. Y. Xian, B. Schiele, Z. Akata. IEEE CVPR 2017.


trainval = r"""002.Laysan_Albatross
003.Sooty_Albatross
015.Lazuli_Bunting
016.Painted_Bunting
020.Yellow_breasted_Chat
022.Chuck_will_Widow
047.American_Goldfinch
048.European_Goldfinch
067.Anna_Hummingbird
068.Ruby_throated_Hummingbird
069.Rufous_Hummingbird
073.Blue_Jay
074.Florida_Jay
075.Green_Jay
076.Dark_eyed_Junco
089.Hooded_Merganser
090.Red_breasted_Merganser
100.Brown_Pelican
149.Brown_Thrasher
150.Sage_Thrasher
001.Black_footed_Albatross
014.Indigo_Bunting
034.Gray_crowned_Rosy_Finch
035.Purple_Finch
101.White_Pelican
120.Fox_Sparrow
110.Geococcyx
085.Horned_Lark
008.Rhinoceros_Auklet
054.Blue_Grosbeak
171.Myrtle_Warbler
041.Scissor_tailed_Flycatcher
056.Pine_Grosbeak
109.American_Redstart
187.American_Three_toed_Woodpecker
175.Pine_Warbler
030.Fish_Crow
195.Carolina_Wren
051.Horned_Grebe
107.Common_Raven
117.Clay_colored_Sparrow
135.Bank_Swallow
134.Cape_Glossy_Starling
046.Gadwall
147.Least_Tern
037.Acadian_Flycatcher
160.Black_throated_Blue_Warbler
126.Nelson_Sharp_tailed_Sparrow
137.Cliff_Swallow
125.Lincoln_Sparrow
027.Shiny_Cowbird
189.Red_bellied_Woodpecker
197.Marsh_Wren
186.Cedar_Waxwing
158.Bay_breasted_Warbler
064.Ring_billed_Gull
044.Frigatebird
007.Parakeet_Auklet
183.Northern_Waterthrush
142.Black_Tern
086.Pacific_Loon
159.Black_and_white_Warbler
081.Pied_Kingfisher
128.Seaside_Sparrow
011.Rusty_Blackbird
145.Elegant_Tern
029.American_Crow
166.Golden_winged_Warbler
059.California_Gull
095.Baltimore_Oriole
155.Warbling_Vireo
010.Red_winged_Blackbird
141.Artic_Tern
102.Western_Wood_Pewee
079.Belted_Kingfisher
106.Horned_Puffin
192.Downy_Woodpecker
005.Crested_Auklet
083.White_breasted_Kingfisher
178.Swainson_Warbler
162.Canada_Warbler
194.Cactus_Wren
196.House_Wren
050.Eared_Grebe
039.Least_Flycatcher
105.Whip_poor_Will
036.Northern_Flicker
032.Mangrove_Cuckoo
146.Forsters_Tern
082.Ringed_Kingfisher
060.Glaucous_winged_Gull
144.Common_Tern
199.Winter_Wren
093.Clark_Nutcracker
198.Rock_Wren
066.Western_Gull
099.Ovenbird
053.Western_Grebe
151.Black_capped_Vireo
018.Spotted_Catbird
152.Blue_headed_Vireo
116.Chipping_Sparrow
061.Heermann_Gull
025.Pelagic_Cormorant
024.Red_faced_Cormorant
078.Gray_Kingbird
017.Cardinal
176.Prairie_Warbler
058.Pigeon_Guillemot
021.Eastern_Towhee
193.Bewick_Wren
057.Rose_breasted_Grosbeak
040.Olive_sided_Flycatcher
153.Philadelphia_Vireo
088.Western_Meadowlark
013.Bobolink
118.House_Sparrow
121.Grasshopper_Sparrow
179.Tennessee_Warbler
062.Herring_Gull
154.Red_eyed_Vireo
092.Nighthawk
038.Great_Crested_Flycatcher
140.Summer_Tanager
182.Yellow_Warbler
096.Hooded_Oriole
172.Nashville_Warbler
071.Long_tailed_Jaeger
042.Vermilion_Flycatcher
185.Bohemian_Waxwing
177.Prothonotary_Warbler
019.Gray_Catbird
065.Slaty_backed_Gull
009.Brewer_Blackbird
112.Great_Grey_Shrike
063.Ivory_Gull
006.Least_Auklet
080.Green_Kingfisher
181.Worm_eating_Warbler
108.White_necked_Raven
122.Harris_Sparrow
115.Brewer_Sparrow
184.Louisiana_Waterthrush
167.Hooded_Warbler
129.Song_Sparrow
133.White_throated_Sparrow
114.Black_throated_Sparrow
170.Mourning_Warbler
131.Vesper_Sparrow
174.Palm_Warbler"""

query = r"""043.Yellow_bellied_Flycatcher
111.Loggerhead_Shrike
023.Brandt_Cormorant
098.Scott_Oriole
055.Evening_Grosbeak
130.Tree_Sparrow
139.Scarlet_Tanager
123.Henslow_Sparrow
156.White_eyed_Vireo
124.Le_Conte_Sparrow
200.Common_Yellowthroat
072.Pomarine_Jaeger
173.Orange_crowned_Warbler
028.Brown_Creeper
119.Field_Sparrow
165.Chestnut_sided_Warbler
103.Sayornis
180.Wilson_Warbler
077.Tropical_Kingbird
012.Yellow_headed_Blackbird
045.Northern_Fulmar
190.Red_cockaded_Woodpecker
191.Red_headed_Woodpecker
138.Tree_Swallow
157.Yellow_throated_Vireo
052.Pied_billed_Grebe
033.Yellow_billed_Cuckoo
164.Cerulean_Warbler
031.Black_billed_Cuckoo
143.Caspian_Tern
094.White_breasted_Nuthatch
070.Green_Violetear
097.Orchard_Oriole
091.Mockingbird
104.American_Pipit
127.Savannah_Sparrow
161.Blue_winged_Warbler
049.Boat_tailed_Grackle
169.Magnolia_Warbler
148.Green_tailed_Towhee
113.Baird_Sparrow
087.Mallard
163.Cape_May_Warbler
136.Barn_Swallow
188.Pileated_Woodpecker
084.Red_legged_Kittiwake
026.Bronzed_Cowbird
004.Groove_billed_Ani
132.White_crowned_Sparrow
168.Kentucky_Warbler"""


trainval = trainval.split("\n")
query = query.split("\n")
